import click
from picomc.globals import gconf


@click.group()
def config_cli():
    """Configure picomc."""
    pass


@config_cli.command()
def show():
    """Print the current config."""
    for k, v in gconf.items():
        print("{}: {}".format(k, v))


@config_cli.command()
@click.argument("key")
@click.argument("value")
def set(key, value):
    gconf[key] = value


@config_cli.command()
@click.argument("key")
def get(key):
    try:
        print(gconf[key])
    except KeyError:
        print("No such attribute.")


@config_cli.command()
@click.argument("key")
def delete(key):
    delattr(gconf, key)


def register_config_cli(picomc_cli):
    picomc_cli.add_command(config_cli, name="config")



import os
import sys
from functools import partial
from os.path import expanduser, join

from picomc.proxy import Proxy

APP_ROOT = {
    "linux": lambda: expanduser("~/.local/share/picomc"),
    "win32": lambda: join(os.getenv("APPDATA"), ".picomc"),
    "darwin": lambda: expanduser("~/Library/Application Support/picomc"),
}[sys.platform]()

try:
    PLATFORM_MAP = {"darwin": "osx", "win32": "windows", "linux": "linux"}
    platform = PLATFORM_MAP[sys.platform]
except KeyError:
    # This is probably not neccesary, as the game is not officialy supported
    # on other platforms and natives are not available. (Unless you compile
    # them and patch the corresponding version json)
    platform = sys.platform


def get_default_java():
    # FIXME. This is just a placeholder.
    return "java"


class Ptr:
    _a = None

    def get(self):
        return self._a

    def set(self, v):
        self._a = v


_ctx_ptr = Ptr()


def _get_object(name):
    ctx = _ctx_ptr.get()
    if ctx is None:
        raise RuntimeError("No context available.")
    r = getattr(ctx, name)
    return r


ctx = Proxy(_ctx_ptr.get)
am = Proxy(partial(_get_object, "am"))
vm = Proxy(partial(_get_object, "vm"))
gconf = Proxy(partial(_get_object, "gconf"))



import hashlib
import json
import os
from functools import partial

from picomc.globals import APP_ROOT
from picomc.logging import logger


class cached_property(object):

    def __init__(self, fn):
        self.fn = fn

    def __get__(self, inst, cls):
        if inst is None:
            return self
        r = self.fn(inst)
        setattr(inst, self.fn.__name__, r)
        return r


def get_filepath(*f):
    return os.path.join(APP_ROOT, *f)


def join_classpath(*cp):
    return os.pathsep.join(cp)


def check_directories():
    """Create directory structure for the application."""
    dirs = [
        "",
        "instances",
        "versions",
        "assets",
        "assets/indexes",
        "assets/objects",
        "assets/virtual",
        "libraries",
    ]
    for d in dirs:
        path = os.path.join(APP_ROOT, *d.split("/"))
        try:
            os.makedirs(path)
            logger.debug("Created dir: {}".format(path))
        except FileExistsError:
            pass


def write_profiles_dummy():
    # This file makes the forge installer happy.
    fname = get_filepath("launcher_profiles.json")
    with open(fname, "w") as fd:
        fd.write(r'{"profiles":{}}')


def file_sha1(filename):
    h = hashlib.sha1()
    with open(filename, "rb", buffering=0) as f:
        for b in iter(partial(f.read, 128 * 1024), b""):
            h.update(b)
    return h.hexdigest()


class ConfigLoader:

    def __init__(self, config_file, defaults={}, dict_impl=dict):
        self.filename = os.path.join(APP_ROOT, config_file)
        self.dict_impl = dict_impl
        self.data = dict_impl(defaults)

    def __enter__(self):
        self._load()
        return self.data

    def __exit__(self, ext_type, exc_value, traceback):
        self._save()

    def _load(self):
        logger.debug("Loading Config from {}.".format(self.filename))
        try:
            with open(self.filename, "r") as json_file:
                self.data.update(
                    json.load(json_file, object_hook=lambda d: self.dict_impl(d))
                )
        except FileNotFoundError:
            pass

    def _save(self):
        logger.debug("Saving Config to {}.".format(self.filename))
        os.makedirs(os.path.dirname(self.filename), exist_ok=True)
        with open(self.filename, "w") as json_file:
            json.dump(self.data, json_file, indent=4)



import logging

logger = logging.getLogger("picomc.cli")


def initialize(debug):
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(levelname)-4s] %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    logger.debug("Logging succesfully initialized.")



class Proxy(object):

    def __init__(self, wrapped):
        self.__dict__["__wrapped__"] = wrapped

    @property
    def __class__(self):
        return self.__wrapped__().__class__

    def __getattr__(self, name):
        return getattr(self.__wrapped__(), name)

    def __setattr__(self, name, value):
        return setattr(self.__wrapped__(), name, value)

    def __delattr__(self, name):
        return delattr(self.__wrapped__(), name)

    def __getitem__(self, name):
        return self.__wrapped__()[name]

    def __setitem__(self, name, value):
        return self.__wrapped__().__setitem__(name, value)

    def __delitem__(self, name):
        return self.__wrapped__().__delitem__(name)



from urllib.parse import urljoin

import requests


class AuthenticationError(Exception):
    pass


class RefreshError(AuthenticationError):
    pass


class MojangYggdrasil:
    BASE_URL = "https://authserver.mojang.com"

    def __init__(self, client_token):
        self.client_token = client_token

    def authenticate(self, username, password):
        ep = urljoin(self.BASE_URL, "/authenticate")
        resp = requests.post(
            ep,
            json={
                "agent": {"name": "Minecraft", "version": 1},
                "username": username,
                "password": password,
                "clientToken": self.client_token,
                "requestUser": True,
            },
        )
        j = resp.json()
        if resp.status_code == 403:
            raise AuthenticationError("Failed to authenticate", j["errorMessage"])
        access_token = j["accessToken"]
        uuid = j["selectedProfile"]["id"]
        name = j["selectedProfile"]["name"]
        return (access_token, uuid, name)

    def refresh(self, access_token):
        ep = urljoin(self.BASE_URL, "/refresh")
        resp = requests.post(
            ep,
            json={
                "accessToken": access_token,
                "clientToken": self.client_token,
                "requestUser": True,
            },
        )
        j = resp.json()
        if resp.status_code == 403:
            raise RefreshError("Failed to refresh", j["errorMessage"])
        access_token = j["accessToken"]
        uuid = j["selectedProfile"]["id"]
        name = j["selectedProfile"]["name"]
        return (access_token, uuid, name)

    def validate(self, access_token):
        ep = urljoin(self.BASE_URL, "/validate")
        resp = requests.post(
            ep, json={"accessToken": access_token, "clientToken": self.client_token}
        )
        return resp.status_code == 204



import json
import operator
import os
import sys
import urllib
from functools import reduce
from platform import architecture

import click
import requests
from picomc.downloader import DownloadQueue
from picomc.globals import platform, vm
from picomc.logging import logger
from picomc.utils import cached_property, file_sha1, get_filepath


class VersionType:
    MAP = {"release": 1, "snapshot": 2, "old_alpha": 4, "old_beta": 8}

    def __init__(self, d=None):
        if isinstance(d, int):
            self._a = d
        elif isinstance(d, str):
            self._a = self.MAP[d]

    def __or__(self, other):
        return VersionType(self._a | other._a)

    def match(self, s):
        return bool(self.MAP[s] & self._a)

    def create(release, snapshot, alpha, beta):
        a = release | (2 * snapshot) | (4 * alpha) | (8 * beta)
        return VersionType(a)


VersionType.RELEASE = VersionType("release")
VersionType.SNAPSHOT = VersionType("snapshot")
VersionType.ALPHA = VersionType("old_alpha")
VersionType.BETA = VersionType("old_beta")

MODE_OVERRIDE = 0
MODE_REDUCE = 1


class CachedVspecAttr(object):

    def __init__(self, attr, mode=MODE_OVERRIDE, reduce_func=None, default=None):
        self.attr = attr
        self.mode = mode
        self.rfunc = reduce_func
        self.default = default

    def __get__(self, vspec, cls):
        if vspec is None:
            return self
        try:
            if self.mode == MODE_OVERRIDE:
                for v in vspec.chain:
                    if self.attr in v.raw_vspec:
                        r = v.raw_vspec[self.attr]
                        break
                else:
                    raise AttributeError()
            elif self.mode == MODE_REDUCE:
                try:
                    r = reduce(
                        self.rfunc,
                        (
                            v.raw_vspec[self.attr]
                            for v in vspec.chain[::-1]
                            if self.attr in v.raw_vspec
                        ),
                    )
                except TypeError as e:
                    raise AttributeError() from e
        except AttributeError as e:
            if self.default:
                r = self.default(vspec.vobj)
        finally:
            try:
                setattr(vspec, self.attr, r)
                return r
            except UnboundLocalError as e:
                raise AttributeError() from e


class VersionSpec:

    def __init__(self, vobj):
        self.vobj = vobj
        self.chain = self._resolve_chain()

    def _resolve_chain(self):
        chain = []
        chain.append(self.vobj)
        cv = self.vobj
        while "inheritsFrom" in cv.raw_vspec:
            cv = Version(cv.raw_vspec["inheritsFrom"])
            chain.append(cv)
        return chain

    minecraftArguments = CachedVspecAttr("minecraftArguments")
    arguments = CachedVspecAttr("arguments")
    mainClass = CachedVspecAttr("mainClass")
    assetIndex = CachedVspecAttr("assetIndex")
    libraries = CachedVspecAttr("libraries", mode=MODE_REDUCE, reduce_func=operator.add)
    jar = CachedVspecAttr("jar", default=lambda vobj: vobj.version_name)
    downloads = CachedVspecAttr("downloads")


class Version:
    LIBRARIES_URL = "https://libraries.minecraft.net/"
    ASSETS_URL = "http://resources.download.minecraft.net/"

    def __init__(self, version_name):
        self.version_name = version_name

    @cached_property
    def jarfile(self):
        v = self.vspec.jar
        return get_filepath("versions", v, "{}.jar".format(v))

    @cached_property
    def raw_vspec(self):
        fpath = get_filepath(
            "versions", self.version_name, "{}.json".format(self.version_name)
        )
        if os.path.exists(fpath):
            with open(fpath) as fp:
                return json.load(fp)
        else:
            url = None
            for v in vm.manifest["versions"]:
                if v["id"] == self.version_name:
                    url = v["url"]
                    break
            if not url:
                raise ValueError("Specified version not avaiable.")
            try:
                j = requests.get(url).json()
                for l in j["libraries"]:
                    try:
                        del l["downloads"]
                    except KeyError:
                        pass
                dirpath = get_filepath("versions", self.version_name)
                os.makedirs(dirpath, exist_ok=True)
                with open(fpath, "w") as fp:
                    json.dump(j, fp, indent=4, sort_keys=True)
                return j
            except requests.ConnectionError:
                logger.error("Failed to retrieve version json file.")
                sys.exit(1)

    @cached_property
    def vspec(self):
        return VersionSpec(self)

    @cached_property
    def raw_asset_index(self):
        iid = self.vspec.assetIndex["id"]
        url = self.vspec.assetIndex["url"]
        fpath = get_filepath("assets", "indexes", "{}.json".format(iid))
        if os.path.exists(fpath):
            with open(fpath) as fp:
                return json.load(fp)
        else:
            try:
                j = requests.get(url).json()
                with open(fpath, "w") as fp:
                    json.dump(j, fp)
                return j
            except requests.ConnectionError:
                logger.error("Failed to retrieve assets index.")
                sys.exit(1)

    def _libraries(self, natives_only=False):
        # The rule matching in this function could be cached,
        # not sure if worth it.
        for lib in self.vspec.libraries:
            rules = lib.get("rules", [])
            allow = "rules" not in lib
            for rule in rules:
                osmatch = True
                if "os" in rule:
                    osmatch = rule["os"]["name"] == platform
                if osmatch:
                    allow = rule["action"] == "allow"
            if not allow:
                continue
            if natives_only and "natives" not in lib:
                continue
            yield lib

    @staticmethod
    def _resolve_library(lib):
        suffix = ""
        if "natives" in lib:
            if platform in lib["natives"]:
                suffix = "-" + lib["natives"][platform]
                # FIXME this is an ugly hack
                suffix = suffix.replace("${arch}", architecture()[0][:2])
            else:
                logger.warn(
                    (
                        "Native library ({}) not available"
                        "for current platform ({}). Ignoring."
                    ).format(lib["name"], platform)
                )
                return None
        fullname = lib["name"]
        url_base = lib.get("url", Version.LIBRARIES_URL)
        p, n, v, *va = fullname.split(":")
        v2 = "-".join([v] + va)

        class LibPaths:
            package = p.replace(".", "/")
            name = n
            version = v
            ext_version = v2
            filename = "{}-{}{}.jar".format(name, ext_version, suffix)
            fullpath = "{}/{}/{}/{}".format(package, name, version, filename)
            url = urllib.parse.urljoin(url_base, fullpath)
            basedir = get_filepath("libraries")
            local_relpath = os.path.join(*fullpath.split("/"))
            local_abspath = os.path.join(basedir, local_relpath)

        return LibPaths

    def lib_filenames(self, natives=False):
        for lib in self._libraries(natives):
            if not natives and "natives" in lib:
                continue
            paths = self._resolve_library(lib)
            if not paths:
                continue
            yield paths.local_abspath

    def download_jarfile(self, force=False):
        """Checks existence and hash of cached jar. Downloads a new one
        if either condition is violated."""
        dlspec = self.vspec.downloads["client"]
        logger.debug("Checking jarfile.")
        if (
            force
            or not os.path.exists(self.jarfile)
            or file_sha1(self.jarfile) != dlspec["sha1"]
        ):
            logger.info("Downloading jar ({}).".format(self.version_name))
            urllib.request.urlretrieve(dlspec["url"], self.jarfile)

    def download_libraries(self, force=False):
        """Downloads missing libraries."""
        logger.info("Checking libraries.")
        q = DownloadQueue()
        for lib in self._libraries():
            paths = self._resolve_library(lib)
            if not paths:
                continue
            if force or not os.path.exists(paths.local_abspath):
                q.add(paths.url, paths.local_relpath)
        q.download(paths.basedir)

    def download_assets(self, force=False):
        """Downloads missing assets."""
        # Produce reverse dict, as multiple files with a single hash
        # can exist and should only be downloaded once.
        rev = dict()
        for name, obj in self.raw_asset_index["objects"].items():
            h = obj["hash"]
            if h in rev:
                rev[h].append(name)
            else:
                rev[h] = [name]

        logger.info(
            "Downloading/checking {} assets. This may take a while.".format(
                len(rev.keys())
            )
        )

        q = DownloadQueue()
        for digest, names in rev.items():
            # This is a mess, should be rewritten. FIXME
            fname = os.path.join("objects", digest[0:2], digest)
            vfnames = (
                os.path.join("virtual", "legacy", *name.split("/")) for name in names
            )
            fullfname = get_filepath("assets", "objects", digest[0:2], digest)
            url = urllib.parse.urljoin(
                self.ASSETS_URL, "{}/{}".format(digest[0:2], digest)
            )
            outs = []
            if force or not os.path.exists(fullfname):
                outs.append(fname)
            if self.raw_asset_index.get("virtual", False):
                for vfname in vfnames:
                    if force or not os.path.exists(
                        get_filepath("assets", *vfname.split("/"))
                    ):
                        outs.append(vfname)
            if outs:
                q.add(url, *outs)

        q.download(get_filepath("assets"))

    def prepare(self):
        self.download_jarfile()
        self.download_libraries()
        self.download_assets()


class VersionManager:
    VERSION_MANIFEST_URL = (
        "https://launchermeta.mojang.com/mc/game/version_manifest.json"
    )

    def resolve_version_name(self, v):
        """Takes a metaversion and resolves to a version."""
        if v == "latest":
            v = self.manifest["latest"]["release"]
            logger.info("Resolved latest -> {}".format(v))
        elif v == "snapshot":
            v = self.manifest["latest"]["snapshot"]
            logger.info("Resolved snapshot -> {}".format(v))
        return v

    @cached_property
    def manifest(self):
        manifest_filepath = get_filepath("versions", "manifest.json")
        try:
            m = requests.get(self.VERSION_MANIFEST_URL).json()
            with open(manifest_filepath, "w") as mfile:
                json.dump(m, mfile, indent=4, sort_keys=True)
            return m
        except requests.ConnectionError:
            logger.warn(
                "Failed to retrieve version_manifest. "
                "Check your internet connection."
            )
            try:
                with open(manifest_filepath) as mfile:
                    logger.warn("Using cached version_manifest.")
                    return json.load(mfile)
            except FileNotFoundError:
                logger.warn("Cached version manifest not available.")
                raise RuntimeError("Failed to retrieve version manifest.")

    def version_list(self, vtype=VersionType.RELEASE, local=False):
        r = [v["id"] for v in self.manifest["versions"] if vtype.match(v["type"])]
        if local:
            import os

            r += (
                "{} [local]".format(name)
                for name in os.listdir(get_filepath("versions"))
                if os.path.isdir(get_filepath("versions", name))
            )
        return r

    def get_version(self, version_name):
        return Version(self.resolve_version_name(version_name))


g_vobj = None


@click.group()
@click.argument("version_name")
def version_cli(version_name):
    """Operate on local Minecraft versions."""
    global g_vobj
    g_vobj = vm.get_version(version_name)


@click.command()
@click.option("--release", is_flag=True, default=False)
@click.option("--snapshot", is_flag=True, default=False)
@click.option("--alpha", is_flag=True, default=False)
@click.option("--beta", is_flag=True, default=False)
@click.option("--local", is_flag=True, default=False)
@click.option("--all", is_flag=True, default=False)
def list_versions(release, snapshot, alpha, beta, local, all):
    """List available Minecraft versions."""
    if all:
        release = snapshot = alpha = beta = local = True
    elif not (release or snapshot or alpha or beta):
        logger.info(
            "Showing only locally installed versions. "
            "Use `version list --help` to get more info."
        )
        local = True
    T = VersionType.create(release, snapshot, alpha, beta)
    print("\n".join(vm.version_list(vtype=T, local=local)))


@version_cli.command()
def prepare():
    g_vobj.prepare()


def register_version_cli(root_cli):
    root_cli.add_command(version_cli, "version")
    root_cli.add_command(list_versions)



import os
import shutil
import subprocess
import zipfile
from platform import architecture

import click
from picomc.account import AccountError
from picomc.globals import am, gconf, platform, vm
from picomc.logging import logger
from picomc.utils import ConfigLoader, get_filepath, join_classpath


class NativesExtractor:

    def __init__(self, instance, vobj):
        self.instance = instance
        self.vobj = vobj
        self.ndir = get_filepath("instances", instance.name, "natives")

    def __enter__(self):
        os.makedirs(self.ndir, exist_ok=True)
        dedup = set()
        for fullpath in self.vobj.lib_filenames(natives=True):
            if fullpath in dedup:
                logger.debug(
                    "Skipping duplicate natives archive: " "{}".format(fullpath)
                )
                continue
            dedup.add(fullpath)
            logger.debug("Extracting natives archive: {}".format(fullpath))
            with zipfile.ZipFile(fullpath) as zf:
                zf.extractall(path=self.ndir)

    def __exit__(self, ext_type, exc_value, traceback):
        logger.debug("Cleaning up natives.")
        shutil.rmtree(self.ndir)
        # print(self.ndir)


def sanitize_name(name):
    return name.replace("..", "_").replace("/", "_")


def process_arguments(arguments_dict):
    """This is a horrible function the only purpose of which is to die and be
    rewritten from scratch. Along with the native library preprocessor."""

    def match_rule(rule):
        # This launcher currently does not support any of the extended
        # features, which currently include at least:
        #   - is_demo_user
        #   - has_custom_resolution
        # It is not clear whether an `os` and `features` matcher may
        # be present simultaneously - assuming not.
        if "features" in rule:
            return False

        osmatch = True
        if "os" in rule:
            # FIXME:
            # The os matcher may apparently also contain a version spec
            # which is probably a regex matched against the java resported
            # os version. See 17w50a.json for an example. Ignoring it for now.
            # This may lead to older versions of Windows matchins as W10.
            # The "name" key is apparently also not required (1.13-pre4.json)
            # and an "arch" rule may be present instead. Possible values are
            # currently unknown.
            if "name" in rule["os"]:
                osmatch = osmatch and rule["os"]["name"] == platform
            if "arch" in rule["os"]:
                logger.warn("Matching arch rule, this may not work.")
                arch = {"32": "x86"}.get(architecture()[0][:2], "?")
                osmatch = osmatch and rule["os"]["arch"] == arch
        if osmatch:
            return rule["action"] == "allow"
        return None

    def subproc(obj):
        args = []
        for a in obj:
            if isinstance(a, str):
                args.append(a)
            else:
                allow = "rules" not in a
                for rule in a["rules"]:
                    m = match_rule(rule)
                    if m is not None:
                        allow = m
                if not allow:
                    continue
                if isinstance(a["value"], list):
                    args.extend(a["value"])
                elif isinstance(a["value"], str):
                    args.append(a["value"])
                else:
                    logger.error("Unknown type of value field.")
        return args

    return (subproc(arguments_dict["game"]), subproc(arguments_dict["jvm"]))


class BackupDict(dict):

    def __getitem__(self, i):
        try:
            return dict.__getitem__(self, i)
        except KeyError:
            return gconf[i]


class InstanceConfigLoader(ConfigLoader):

    def __init__(self, instance_name):
        default_config = {"version": "latest"}
        cfg_file = os.path.join("instances", instance_name, "config.json")
        ConfigLoader.__init__(self, cfg_file, default_config, dict_impl=BackupDict)


class Instance:

    def __init__(self, name):
        self.name = sanitize_name(name)

    def __enter__(self):
        self._cl = InstanceConfigLoader(self.name)
        self.config = self._cl.__enter__()
        return self

    def __exit__(self, ext_type, exc_value, traceback):
        self._cl.__exit__(ext_type, exc_value, traceback)
        del self.config
        del self._cl

    def get_java(self):
        return self.config["java.path"]

    def populate(self, version):
        self.config["version"] = version

    def launch(self, account, version):
        vobj = vm.get_version(version or self.config["version"])
        logger.info("Launching instance {}!".format(self.name))
        logger.info("Using minecraft version: {}".format(vobj.version_name))
        vobj.prepare()
        logger.info("Using account: {}".format(account))
        os.makedirs(get_filepath("instances", self.name, "minecraft"), exist_ok=True)
        with NativesExtractor(self, vobj):
            self._exec_mc(account, vobj)

    def _exec_mc(self, account, v):
        # this is temporary. FIXME
        # This 'function' is quickly getting worse and worse.
        # Rewrite it.

        java = [self.get_java()]
        java.append("-Xms{}".format(self.config["java.memory.min"]))
        java.append("-Xmx{}".format(self.config["java.memory.max"]))
        java += self.config["java.jvmargs"].split()
        libs = list(v.lib_filenames())
        libs.append(v.jarfile)
        classpath = join_classpath(*libs)

        version_type, user_type = (
            ("picomc", "mojang") if account.online else ("picomc/offline", "offline")
        )

        # Make functions out of these two
        natives = get_filepath("instances", self.name, "natives")
        gamedir = get_filepath("instances", self.name, "minecraft")

        mc = v.vspec.mainClass

        if hasattr(v.vspec, "minecraftArguments"):
            mcargs = v.vspec.minecraftArguments.split()
            sjvmargs = ["-Djava.library.path={}".format(natives), "-cp", classpath]
        elif hasattr(v.vspec, "arguments"):
            mcargs, jvmargs = process_arguments(v.vspec.arguments)
            sjvmargs = []
            for a in jvmargs:
                a = a.replace("${", "{")
                a = a.format(
                    natives_directory=natives,
                    launcher_name="picomc",
                    launcher_version="0",  # Do something proper here. FIXME.
                    classpath=classpath,
                )
                sjvmargs.append(a)

        account.refresh()

        smcargs = []
        for a in mcargs:
            # This should be done differently.
            a = a.replace("${", "{")
            a = a.format(
                auth_player_name=account.gname,
                auth_uuid=account.uuid,
                auth_access_token=account.access_token,
                # Only used in old versions.
                auth_session="token:{}:{}".format(account.access_token, account.uuid),
                user_type=user_type,
                user_properties={},
                version_type=version_type,
                version_name=v.version_name,
                game_directory=gamedir,
                assets_root=get_filepath("assets"),
                assets_index_name=v.vspec.assetIndex["id"],
                # FIXME Ugly hack relying on untested behaviour:
                game_assets=get_filepath("assets", "virtual", "legacy"),
            )
            smcargs.append(a)

        fargs = java + sjvmargs + [mc] + smcargs
        logger.debug("Launching: " + " ".join(fargs))
        subprocess.run(fargs, cwd=gamedir)

    @staticmethod
    def exists(name):
        name = sanitize_name(name)
        return os.path.exists(get_filepath("instances", name, "config.json"))

    @staticmethod
    def remove(name):
        shutil.rmtree(get_filepath("instances", name))


def instance_list():
    import os

    yield from (
        name for name in os.listdir(get_filepath("instances")) if Instance.exists(name)
    )


@click.group()
@click.argument("instance_name")
def instance_cli(instance_name):
    """Manage your instances."""
    instance_name = sanitize_name(instance_name)
    global g_iname
    g_iname = instance_name


@click.command()
@click.argument("instance_name")
@click.argument("version", default="latest")
def create_instance(instance_name, version):
    """Create a new instance."""
    if Instance.exists(instance_name):
        logger.error("An instance with that name already exists.")
        return
    with Instance(instance_name) as inst:
        inst.populate(version)


@click.command()
def list_instances():
    """Show a list of instances."""
    print("\n".join(instance_list()))


@instance_cli.command()
def remove():
    if Instance.exists(g_iname):
        Instance.remove(g_iname)
    else:
        logger.error("No such instance exists.")


@instance_cli.command()
@click.option("--account", default=None)
@click.option("--version-override", default=None)
def launch(account, version_override):
    """Launch the instance."""
    if account is None:
        account = am.get_default()
    else:
        account = am.get(account)
    if not Instance.exists(g_iname):
        logger.error("No such instance exists.")
        return
    with Instance(g_iname) as inst:
        try:
            inst.launch(account, version_override)
        except AccountError as e:
            logger.error("Not launching due to account error: {}".format(e))


@instance_cli.command()
def dir():
    """Print root directory of instance."""
    if not g_iname:
        print(get_filepath("instances"))
    else:
        # Careful, if configurable instance dirs are added, this breaks.
        print(get_filepath("instances", g_iname))


def register_instance_cli(picomc_cli):
    picomc_cli.add_command(instance_cli, name="instance")
    picomc_cli.add_command(create_instance)
    picomc_cli.add_command(list_instances)



from picomc import __version__
from setuptools import find_packages, setup

setup(
    name="picomc",
    version=__version__,
    description="A very small CLI Minecraft launcher.",
    url="https://git.sammserver.com/sammko/picomc",
    author="Samuel Čavoj",
    author_email="sammko@sammserver.com",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    packages=find_packages(),
    install_requires=["click", "requests"],
    python_requires=">=3",
    entry_points={"console_scripts": ["picomc = picomc:main"]},
)



import uuid

import click
from picomc.globals import am
from picomc.logging import logger
from picomc.utils import ConfigLoader
from picomc.yggdrasil import AuthenticationError, MojangYggdrasil, RefreshError


class NAMESPACE_NULL:
    bytes = b""


def generate_client_token():
    # Any random string, this matches the behaviour of the official launcher.
    return str(uuid.uuid4().hex)


class Account:

    def __init__(self, **kwargs):
        self.__dict__.update(self.DEFAULTS)
        self.__dict__.update(kwargs)

    def __repr__(self):
        return self.name

    def to_dict(self):
        return {k: getattr(self, k) for k in self.DEFAULTS.keys()}

    @classmethod
    def from_config(cls, name, config):
        c = OnlineAccount if config.get("online", False) else OfflineAccount
        return c(name=name, **config)


class OfflineAccount(Account):
    DEFAULTS = {"uuid": "-", "online": False}
    access_token = "-"

    @classmethod
    def new(cls, name):
        u = uuid.uuid3(NAMESPACE_NULL, "OfflinePlayer:{}".format(name)).hex
        return cls(name=name, uuid=u)

    @property
    def gname(self):
        return self.name

    def refresh(self):
        return False


class OnlineAccount(Account):
    DEFAULTS = {
        "uuid": "-",
        "online": True,
        "gname": "-",
        "access_token": "-",
        "is_authenticated": False,
        "username": "-",
    }

    fresh = False

    @classmethod
    def new(cls, name, username):
        return cls(name=name, username=username)

    def validate(self):
        r = am.yggdrasil.validate(self.access_token)
        if r:
            self.fresh = True
        return r

    def refresh(self, force=False):
        if self.fresh and not force:
            return False
        if self.is_authenticated:
            if self.validate():
                return
            else:
                try:
                    refresh = am.yggdrasil.refresh(self.access_token)
                    self.access_token, self.uuid, self.gname = refresh
                    self.fresh = True
                    return True
                except RefreshError as e:
                    logger.error(
                        "Failed to refresh access_token," " please authenticate again."
                    )
                    self.is_authenticated = False
                    raise e
                finally:
                    am.save(self)
        else:
            raise AccountError("Not authenticated.")

    def authenticate(self, password):
        self.access_token, self.uuid, self.gname = am.yggdrasil.authenticate(
            self.username, password
        )
        self.is_authenticated = True
        self.fresh = True
        am.save(self)


class AccountError(ValueError):

    def __str__(self):
        return " ".join(self.args)


DEFAULT_CONFIG = {
    "default": None,
    "accounts": {},
    "client_token": generate_client_token(),
}


class AccountManager:
    cfg_file = "accounts.json"

    def __enter__(self):
        self._cl = ConfigLoader(self.cfg_file, DEFAULT_CONFIG)
        self.config = self._cl.__enter__()
        self.yggdrasil = MojangYggdrasil(self.config["client_token"])
        return self

    def __exit__(self, ext_type, exc_value, traceback):
        self._cl.__exit__(ext_type, exc_value, traceback)
        del self.config
        del self._cl

    def list(self):
        return self.config["accounts"].keys()

    def get(self, name):
        try:
            acc = Account.from_config(name, self.config["accounts"][name])
            acc.is_default = self.config["default"] == name
            return acc
        except KeyError as ke:
            raise AccountError("Account does not exist:", name) from ke

    def exists(self, name):
        return name in self.config["accounts"]

    def get_default(self):
        default = self.config["default"]
        if not default:
            raise AccountError("Default account not configured.")
        return self.get(default)

    def is_default(self, name):
        return name == self.config["default"]

    def set_default(self, account):
        self.config["default"] = account.name

    def add(self, account):
        if am.exists(account.name):
            raise AccountError("An account already exists with that name.")
        if not self.config["default"] and not self.config["accounts"]:
            self.config["default"] = account.name
        self.save(account)

    def save(self, account):
        self.config["accounts"][account.name] = account.to_dict()

    def remove(self, name):
        try:
            if self.config["default"] == name:
                self.config["default"] = None
            del self.config["accounts"][name]
        except KeyError:
            raise AccountError("Account does not exist:", name)


g_aname = None


@click.group()
@click.argument("account_name")
def account_cli(account_name):
    """Manage your accounts."""
    global g_aname
    g_aname = account_name


@click.command()
def list_accounts():
    """List avaiable accounts."""
    alist = am.list()
    if alist:
        print(
            "\n".join(
                "{}{}".format("* " if am.is_default(u) else "  ", u) for u in alist
            )
        )
    else:
        print("No accounts.")


@click.command()
@click.argument("account_name")
@click.argument("mojang_username", default="")
def create_account(account_name, mojang_username):
    """Create an account."""
    try:
        if mojang_username:
            acc = OnlineAccount.new(account_name, mojang_username)
        else:
            acc = OfflineAccount.new(account_name)
        am.add(acc)
    except AccountError as e:
        print(e)


@account_cli.command()
def authenticate():
    import getpass

    try:
        a = am.get(g_aname)
        # add some output here
        p = getpass.getpass("Password: ")
        a.authenticate(p)
    except AuthenticationError as e:
        print(e)


@account_cli.command()
def refresh():
    try:
        a = am.get(g_aname)
        a.refresh()
    except (AccountError, RefreshError) as e:
        print(e)


@account_cli.command()
def remove():
    try:
        am.remove(g_aname)
    except AccountError as e:
        print(e)


@account_cli.command()
def setdefault():
    try:
        default = am.get(g_aname)
        am.set_default(default)
    except AccountError as e:
        print(e)


def register_account_cli(picomc_cli):
    picomc_cli.add_command(account_cli, name="account")
    picomc_cli.add_command(create_account)
    picomc_cli.add_command(list_accounts)



from contextlib import ExitStack

import click
import picomc.logging
from picomc.account import AccountManager, register_account_cli
from picomc.config import register_config_cli
from picomc.globals import APP_ROOT, _ctx_ptr, ctx
from picomc.instances import register_instance_cli
from picomc.logging import logger
from picomc.utils import ConfigLoader, check_directories, write_profiles_dummy
from picomc.version import VersionManager, register_version_cli

__version__ = "0.1.5"


@click.group()
@click.option("--debug/--no-debug", default=False)
@click.option("-r", "--root", help="Application data directory.", default=APP_ROOT)
@click.version_option(version=__version__, prog_name="picomc")
@click.pass_obj
def picomc_cli(es, debug, root):
    """picomc is a minimal CLI Minecraft launcher."""
    picomc.logging.initialize(debug)
    picomc.globals.APP_ROOT = root
    check_directories()

    write_profiles_dummy()

    logger.debug("Using application directory: {}".format(picomc.globals.APP_ROOT))

    am = es.enter_context(AccountManager())
    ctx.am = am
    default_config = {
        "java.path": "java",
        "java.memory.min": "128M",
        "java.memory.max": "1G",
        "java.jvmargs": "",
    }
    gconf = es.enter_context(ConfigLoader("config.json", defaults=default_config))
    ctx.gconf = gconf
    ctx.vm = VersionManager()


def main():

    class Context:
        pass

    c = Context()
    _ctx_ptr.set(c)
    with ExitStack() as estack:
        picomc_cli(obj=estack)


register_account_cli(picomc_cli)
register_version_cli(picomc_cli)
register_instance_cli(picomc_cli)
register_config_cli(picomc_cli)



import os
import shutil
import subprocess
import urllib

from picomc.logging import logger


def check_aria2():
    try:
        subprocess.run(
            "aria2c --version".split(),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        return True
    except FileNotFoundError:
        return False


def downloader_aria2(q, d):
    p = subprocess.Popen(
        "/usr/bin/aria2c -i - -j8 --dir".split() + [d],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    L = []
    for url, outs in q:
        L.append("{}\n out={}".format(url, outs[0]))
    out, err = p.communicate("\n".join(L).encode())
    if p.returncode:
        print(out.decode(), err.decode())
        raise RuntimeError("Failed to download files.")
    else:
        for url, outs in q:
            src = os.path.join(d, outs[0])
            for o in outs[1:]:
                dst = os.path.join(d, o)
                os.makedirs(os.path.dirname(dst), exist_ok=True)
                shutil.copy(src, dst)


def downloader_urllib(q, d):
    # Dumb and slow.
    for url, outs in q:
        aouts = list(os.path.join(d, o) for o in outs)
        for o in aouts:
            os.makedirs(os.path.dirname(o), exist_ok=True)
        pout, *eouts = aouts
        urllib.request.urlretrieve(url, pout)
        for o in eouts:
            shutil.copy(pout, o)


class DownloadQueue:

    def __init__(self):
        self.q = []

    def add(self, url, *filename):
        self.q.append((url, filename))

    def download(self, d):
        if not self.q:
            return
        if check_aria2():
            logger.info("Using aria2 downloader.")
            downloader = downloader_aria2
        else:
            logger.info("Using urllib downloader.")
            downloader = downloader_urllib
        downloader(self.q, d)
